import os.path
import tempfile

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
DEBUG = True

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}
INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.staticfiles',

    'storages',
    's3file',
    'tests.testapp',
)

DEFAULT_FILE_STORAGE = 'tests.testapp.dummy_storage.DummyS3Boto3Storage'

MIDDLEWARE = MIDDLEWARE_CLASSES = (
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    's3file.middleware.S3FileMiddleware',
)


SECRET_KEY = 'SuperSecretKey'
STATIC_URL = '/static/'
MEDIA_ROOT = tempfile.mkdtemp()
ROOT_URLCONF = 'tests.testapp.urls'
MEDIA_URL = '/s3/'
SITE_ID = 1

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'APP_DIRS': True,
    },
]

USE_L10N = True

AWS_ACCESS_KEY_ID = 'testaccessid'
AWS_SECRET_ACCESS_KEY = 'supersecretkey'
AWS_STORAGE_BUCKET_NAME = 'test-bucket'
AWS_S3_REGION_NAME = 'eu-central-1'
AWS_S3_SIGNATURE_VERSION = 's3v4'



import os
import tempfile

import pytest
from django.core.files.base import ContentFile
from django.utils.encoding import force_text
from selenium import webdriver
from selenium.common.exceptions import WebDriverException


@pytest.yield_fixture(scope='session')
def driver():
    chrome_options = webdriver.ChromeOptions()
    chrome_options.add_argument('headless')
    chrome_options.add_argument('window-size=1200x800')
    try:
        b = webdriver.Chrome(chrome_options=chrome_options)
    except WebDriverException as e:
        pytest.skip(force_text(e))
    else:
        yield b
        b.quit()


@pytest.fixture
def upload_file(request):
    path = tempfile.mkdtemp()
    file_name = os.path.join(path, '%s.txt' % request.node.name)
    with open(file_name, 'w') as f:
        f.write(request.node.name)
    return file_name


@pytest.fixture
def filemodel(request, db):
    from tests.testapp.models import FileModel

    return FileModel.objects.create(
        file=ContentFile(request.node.name, '%s.txt' % request.node.name)
    )



from django.core.files.base import ContentFile
from django.core.files.storage import default_storage
from django.core.files.uploadedfile import SimpleUploadedFile

from s3file.middleware import S3FileMiddleware


class TestS3FileMiddleware:

    def test_get_files_from_storage(self):
        content = b'test_get_files_from_storage'
        default_storage.save('test_get_files_from_storage', ContentFile(content))
        files = S3FileMiddleware.get_files_from_storage(['test_get_files_from_storage'])
        file = next(files)
        assert file.read() == content

    def test_process_request(self, rf):
        uploaded_file = SimpleUploadedFile('uploaded_file.txt', b'uploaded')
        request = rf.post('/', data={'file': uploaded_file})
        S3FileMiddleware().process_request(request)
        assert request.FILES.getlist('file')
        assert request.FILES.get('file').read() == b'uploaded'

        default_storage.save('s3_file.txt', ContentFile(b's3file'))
        request = rf.post('/', data={'file': 's3_file.txt', 's3file': 'file'})
        S3FileMiddleware().process_request(request)
        assert request.FILES.getlist('file')
        assert request.FILES.get('file').read() == b's3file'



from django.apps import AppConfig


class S3FileConfig(AppConfig):
    name = 's3file'
    verbose_name = 'S3File'

    def ready(self):
        from django.core.files.storage import default_storage
        from storages.backends.s3boto3 import S3Boto3Storage

        if isinstance(default_storage, S3Boto3Storage):
            from django import forms
            from .forms import S3FileInputMixin

            if S3FileInputMixin not in forms.ClearableFileInput.__bases__:
                forms.ClearableFileInput.__bases__ = \
                    (S3FileInputMixin,) + forms.ClearableFileInput.__bases__



from django.conf.urls import url
from django.views.generic import FormView

from . import forms, views

urlpatterns = [
    url(r'^s3/$',
        views.S3MockView.as_view(), name='s3mock'),
    url(r'^upload/$',
        views.ExampleFormView.as_view(), name='upload'),
]



from django.core.files.storage import default_storage
from django.http import JsonResponse, response
from django.views import generic

from tests.testapp import forms


class S3MockView(generic.View):

    def post(self, request):
        file = request.FILES.get('file', None)
        key = request.POST.get('key')
        key = key.replace('${filename}', file.name)
        default_storage.save(key, file)
        return response.HttpResponse(
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<PostResponse>'
            '<Location></Location>'
            '<Bucket>AWS_STORAGE_BUCKET_NAME</Bucket>'
            '<Key>%s</Key>'
            '<ETag>"1e2580c388265551922a1f73ae5954a3"</ETag>'
            '</PostResponse>' % key,
            status=201)


class ExampleFormView(generic.FormView):
    form_class = forms.UploadForm
    template_name = 'form.html'

    def form_valid(self, form):
        return JsonResponse(self.request.POST, status=201)



import importlib

from django import forms

from s3file.apps import S3FileConfig
from s3file.forms import S3FileInputMixin


class TestS3FileConfig:
    def test_ready(self, settings):
        app = S3FileConfig('s3file', importlib.import_module('tests.testapp'))
        app.ready()
        assert not isinstance(forms.ClearableFileInput(), S3FileInputMixin)
        settings.DEFAULT_FILE_STORAGE = 'storages.backends.s3boto3.S3Boto3Storage'
        app.ready()
        assert isinstance(forms.ClearableFileInput(), S3FileInputMixin)



import logging
import os
import uuid

from django.conf import settings
from django.core.files.storage import default_storage
from django.forms.widgets import ClearableFileInput
from django.utils.functional import cached_property

logger = logging.getLogger('s3file')


class S3FileInputMixin:
    """FileInput that uses JavaScript to directly upload to Amazon S3."""

    needs_multipart_form = False
    upload_path = getattr(settings, 'S3FILE_UPLOAD_PATH', os.path.join('tmp', 's3file'))
    expires = settings.SESSION_COOKIE_AGE

    @property
    def bucket_name(self):
        return default_storage.bucket.name

    @property
    def client(self):
        return default_storage.connection.meta.client

    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)

        accept = attrs.get('accept')
        response = self.client.generate_presigned_post(
            self.bucket_name, os.path.join(self.upload_folder, '${filename}'),
            Conditions=self.get_conditions(accept),
            ExpiresIn=self.expires,
        )

        defaults = {
            'data-fields-%s' % key: value
            for key, value in response['fields'].items()
        }
        defaults['data-url'] = response['url']
        defaults.update(attrs)

        try:
            defaults['class'] += ' s3file'
        except KeyError:
            defaults['class'] = 's3file'
        return defaults

    def get_conditions(self, accept):
        conditions = [
            {"bucket": self.bucket_name},
            ["starts-with", "$key", self.upload_folder],
            {"success_action_status": "201"},
        ]
        if accept:
            accept = accept.replace(' ', '')  # remove whitespaces
            mime_types = accept.split(',') if accept else []  # catch empty string
            for mime_type in mime_types:
                top_type, sub_type = mime_type.split('/', 1)
                if sub_type == '*':
                    conditions.append(["starts-with", "$Content-Type", "%s/" % top_type])
                else:
                    conditions.append({"Content-Type": mime_type})
        else:
            conditions.append(["starts-with", "$Content-Type", ""])

        return conditions

    @cached_property
    def upload_folder(self):
        return os.path.join(
            self.upload_path,
            uuid.uuid4().hex,
        )

    class Media:
        js = (
            's3file/js/s3file.js',
        )



import os

from django.core.files.storage import default_storage

try:
    from django.utils.deprecation import MiddlewareMixin
except ImportError:
    MiddlewareMixin = object


class S3FileMiddleware(MiddlewareMixin):
    @staticmethod
    def get_files_from_storage(paths):
        """Return S3 file where the name does not include the path."""
        for path in paths:
            f = default_storage.open(path)
            f.name = os.path.basename(path)
            yield f

    def process_request(self, request):
        file_fields = request.POST.getlist('s3file', [])
        for field_name in file_fields:
            paths = request.POST.getlist(field_name, [])
            request.FILES.setlist(field_name, list(self.get_files_from_storage(paths)))



import json
from contextlib import contextmanager

import pytest
from django.core.files.storage import default_storage
from django.forms import ClearableFileInput
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.support.expected_conditions import staleness_of
from selenium.webdriver.support.wait import WebDriverWait

from s3file.forms import S3FileInputMixin
from tests.testapp.forms import UploadForm

try:
    from django.urls import reverse
except ImportError:
    # Django 1.8 support
    from django.core.urlresolvers import reverse


@contextmanager
def wait_for_page_load(driver, timeout=30):
    old_page = driver.find_element_by_tag_name('html')
    yield
    WebDriverWait(driver, timeout).until(
        staleness_of(old_page)
    )


class TestS3FileInput:
    @property
    def url(self):
        return reverse('upload')

    @pytest.fixture(autouse=True)
    def patch(self):
        if S3FileInputMixin not in ClearableFileInput.__bases__:
            ClearableFileInput.__bases__ = \
                (S3FileInputMixin,) + ClearableFileInput.__bases__
        pass

    @pytest.fixture
    def freeze(self, monkeypatch):
        """Freeze datetime and UUID."""
        monkeypatch.setattr('s3file.forms.S3FileInputMixin.upload_folder', 'tmp')

    def test_value_from_datadict(self, client, upload_file):
        with open(upload_file) as f:
            uploaded_file = default_storage.save('test.jpg', f)
        response = client.post(reverse('upload'), {
            'file': uploaded_file,
            's3file': 'file'
        })

        assert response.status_code == 201

    def test_value_from_datadict_initial_data(self, filemodel):
        form = UploadForm(instance=filemodel)
        assert filemodel.file.name in form.as_p(), form.as_p()
        assert not form.is_valid()

    def test_file_does_not_exist_no_fallback(self, filemodel):
        form = UploadForm(data={'file': 'foo.bar', 's3file': 'file'}, instance=filemodel)
        assert form.is_valid()
        assert form.cleaned_data['file'] == filemodel.file

    def test_initial_no_file_uploaded(self, filemodel):
        form = UploadForm(data={'file': ''}, instance=filemodel)
        assert form.is_valid(), form.errors
        assert not form.has_changed()
        assert form.cleaned_data['file'] == filemodel.file

    def test_initial_fallback(self, filemodel):
        form = UploadForm(data={'file': ''}, instance=filemodel)
        assert form.is_valid()
        assert form.cleaned_data['file'] == filemodel.file

    def test_clear(self, filemodel):
        form = UploadForm(data={'file-clear': '1'}, instance=filemodel)
        assert form.is_valid()
        assert not form.cleaned_data['file']

    def test_build_attr(self):
        assert set(ClearableFileInput().build_attrs({}).keys()) == {
            'class',
            'data-url',
            'data-fields-x-amz-algorithm',
            'data-fields-x-amz-date',
            'data-fields-x-amz-signature',
            'data-fields-x-amz-credential',
            'data-fields-policy',
            'data-fields-key',
        }
        assert ClearableFileInput().build_attrs({})['class'] == 's3file'
        assert ClearableFileInput().build_attrs(
            {'class': 'my-class'})['class'] == 'my-class s3file'

    def test_get_conditions(self, freeze):
        conditions = ClearableFileInput().get_conditions(None)
        assert all(condition in conditions for condition in [
            {"bucket": 'test-bucket'},
            {"success_action_status": "201"},
            ['starts-with', '$key', 'tmp'],
            ["starts-with", "$Content-Type", ""]
        ]), conditions

    def test_accept(self):
        widget = ClearableFileInput()
        assert 'accept' not in widget.render(name='file', value='test.jpg')
        assert ["starts-with", "$Content-Type", ""] in widget.get_conditions(None)

        widget = ClearableFileInput(attrs={'accept': 'image/*'})
        assert 'accept="image/*"' in widget.render(name='file', value='test.jpg')
        assert ["starts-with", "$Content-Type", "image/"] in widget.get_conditions('image/*')

        widget = ClearableFileInput(attrs={'accept': 'image/jpeg'})
        assert 'accept="image/jpeg"' in widget.render(name='file', value='test.jpg')
        assert {"Content-Type": 'image/jpeg'} in widget.get_conditions('image/jpeg')

        widget = ClearableFileInput(attrs={'accept': 'application/pdf,image/*'})
        assert 'accept="application/pdf,image/*"' in widget.render(name='file', value='test.jpg')
        assert ["starts-with", "$Content-Type", "image/"] in widget.get_conditions(
            'application/pdf,image/*')
        assert {"Content-Type": 'application/pdf'} in widget.get_conditions(
            'application/pdf,image/*')

        widget = ClearableFileInput(attrs={'accept': 'application/pdf, image/*'})
        assert 'accept="application/pdf, image/*"' in widget.render(name='file', value='test.jpg')
        assert ["starts-with", "$Content-Type", "image/"] in widget.get_conditions(
            'application/pdf, image/*')
        assert {"Content-Type": 'application/pdf'} in widget.get_conditions(
            'application/pdf, image/*')

    def test_no_js_error(self, driver, live_server):
        driver.get(live_server + self.url)

        with pytest.raises(NoSuchElementException):
            error = driver.find_element_by_xpath('//body[@JSError]')
            pytest.fail(error.get_attribute('JSError'))

    def test_file_insert(self, request, driver, live_server, upload_file, freeze):
        driver.get(live_server + self.url)
        file_input = driver.find_element_by_xpath('//input[@type=\'file\']')
        file_input.send_keys(upload_file)
        assert file_input.get_attribute('name') == 'file'
        with wait_for_page_load(driver, timeout=10):
            file_input.submit()
        assert default_storage.exists('tmp/%s.txt' % request.node.name)

        with pytest.raises(NoSuchElementException):
            error = driver.find_element_by_xpath('//body[@JSError]')
            pytest.fail(error.get_attribute('JSError'))

    def test_file_insert_submit_value(self, driver, live_server, upload_file, freeze):
        driver.get(live_server + self.url)
        file_input = driver.find_element_by_xpath('//input[@type=\'file\']')
        file_input.send_keys(upload_file)
        assert file_input.get_attribute('name') == 'file'
        save_button = driver.find_element_by_xpath('//input[@name=\'save\']')
        with wait_for_page_load(driver, timeout=10):
            save_button.click()
        assert 'save' in driver.page_source

        driver.get(live_server + self.url)
        file_input = driver.find_element_by_xpath('//input[@type=\'file\']')
        file_input.send_keys(upload_file)
        assert file_input.get_attribute('name') == 'file'
        save_button = driver.find_element_by_xpath('//button[@name=\'save_continue\']')
        with wait_for_page_load(driver, timeout=10):
            save_button.click()
        assert 'save_continue' in driver.page_source
        assert 'continue_value' in driver.page_source

    def test_media(self):
        assert ClearableFileInput().media._js == ['s3file/js/s3file.js']



from django.core.files.storage import FileSystemStorage

try:
    from django.urls import reverse
except ImportError:
    # Django 1.8 support
    from django.core.urlresolvers import reverse


class DummyS3Boto3Storage(FileSystemStorage):
    class connection:
        class meta:
            class client:
                @staticmethod
                def generate_presigned_post(*args, **kargs):
                    return {
                        'url': reverse('s3mock'),
                        'fields': {
                            'x-amz-algorithm': 'AWS4-HMAC-SHA256',
                            'x-amz-date': '20170908T111600Z',
                            'x-amz-signature': 'asdf',
                            'x-amz-credential': 'testaccessid',
                            'policy': 'asdf',
                            'key': 'tmp/${filename}',
                        },
                    }

    class bucket:
        name = 'test-bucket'
